/*
  audio.wasapi (2011-08-16)
  author: Fatbag
*/

#include <Windows.h>
#include "wasapi.hpp"

namespace ruby {

class pAudioWASAPI {
public:
  IMMDevice *pDevice;
  IAudioClient *pAudioClient;
  IAudioRenderClient *pRenderClient;
  HANDLE hEvent, BufferAvailable, hThread;
  volatile uint32_t Terminate;

  struct {
    uint64_t latency;
    volatile unsigned bufferframes;

    uint32_t volatile *localbuffer;
    unsigned localbuffers;
    volatile uint32_t buffersfilled;
    unsigned onbuffer;
    unsigned frameoffset;
  } device;

  struct {
    bool synchronize;
    unsigned frequency;
    unsigned latency;
  } settings;

  bool cap(const string& name) {
    if(name == Audio::Synchronize) return true;
    if(name == Audio::Frequency) return true;
    if(name == Audio::Latency) return true;
    return false;
  }

  any get(const string& name) {
    if(name == Audio::Synchronize) return settings.synchronize;
    if(name == Audio::Frequency) return settings.frequency;
    if(name == Audio::Latency) return device.localbuffers*device.latency/1000;
    return false;
  }

  bool set(const string& name, const any& value) {
    if(name == Audio::Synchronize) {
      settings.synchronize = any_cast<bool>(value);
      return true;
    }

    if(name == Audio::Frequency) {
      settings.frequency = any_cast<unsigned>(value);
      return true;
    }

    if(name == Audio::Latency) {
      settings.latency = any_cast<unsigned>(value);
      return true;
    }

    return false;
  }

  void sample(uint16_t left, uint16_t right) {
    uint32_t thisframe = left | (right << 16);

    if(settings.synchronize) {
      WaitForSingleObject(BufferAvailable, INFINITE);
    } else if(WaitForSingleObject(BufferAvailable, 0)) return; //Discard if no buffers available

    device.localbuffer[device.onbuffer*device.bufferframes + device.frameoffset++] = thisframe;

    if(device.frameoffset == device.bufferframes) {
      InterlockedIncrement((volatile LONG*) &device.buffersfilled);
      if(device.buffersfilled == device.localbuffers) {
        ResetEvent(BufferAvailable);
      }

      device.frameoffset = 0;
      device.onbuffer++; if(device.onbuffer == device.localbuffers) device.onbuffer = 0;
    }
  }

  void clear() {
  }

  bool init() {
    if(!hEvent) hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    if(!BufferAvailable) BufferAvailable = CreateEvent(NULL, TRUE, TRUE, NULL);

    hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) ThreadProc, this, 0, NULL);
    if(hThread == NULL) {
      return false;
    }

    WaitForSingleObject(hEvent, INFINITE);
    if(Terminate == 1) {
      return false;
    }

    timeBeginPeriod(1);

    return true;
  }

  static bool threadinit(pAudioWASAPI& obj, HANDLE hEvent) {
    CoInitializeEx(NULL, COINIT_MULTITHREADED | COINIT_SPEED_OVER_MEMORY);

    IMMDeviceEnumerator *pEnumerator;
    if(FAILED(CoCreateInstance(CLSID_MMDeviceEnumerator, NULL, CLSCTX_ALL, IID_IMMDeviceEnumerator, (void**)&pEnumerator))) {
      return false;
    }
    pEnumerator->GetDefaultAudioEndpoint(eRender, eConsole, &obj.pDevice);
    pEnumerator->Release();
    pEnumerator = 0;
    if(!obj.pDevice) {
      return false;
    }

    if(FAILED(obj.pDevice->Activate(IID_IAudioClient, CLSCTX_ALL, NULL, (void**)&obj.pAudioClient))) {
      return false;
    }

    WAVEFORMATEX wfx;
    wfx.wFormatTag = WAVE_FORMAT_PCM;
    wfx.nChannels       = 2;
    wfx.wBitsPerSample  = 16;
    wfx.nBlockAlign     = 4;
    wfx.nSamplesPerSec  = obj.settings.frequency;
    wfx.nAvgBytesPerSec = obj.settings.frequency*4;
    wfx.cbSize = 0;
    if(obj.pAudioClient->IsFormatSupported(AUDCLNT_SHAREMODE_EXCLUSIVE, &wfx, NULL) == S_FALSE) {
      return false;
    }

    obj.pAudioClient->GetDevicePeriod(NULL, (LONGLONG*) &obj.device.latency);
    if(FAILED(obj.pAudioClient->Initialize(AUDCLNT_SHAREMODE_EXCLUSIVE, AUDCLNT_STREAMFLAGS_EVENTCALLBACK,
      obj.device.latency, obj.device.latency, &wfx, NULL))) {
      return false;
    }
    obj.device.localbuffers = (2000*obj.settings.latency + obj.device.latency)/(2*obj.device.latency);
    obj.pAudioClient->SetEventHandle(hEvent);

    uint32_t bufferframes;
    obj.pAudioClient->GetBufferSize(&bufferframes);
    InterlockedExchange((volatile LONG*) &obj.device.bufferframes, bufferframes);
    InterlockedExchangePointer((void* volatile*) &obj.device.localbuffer,
      new uint32_t[obj.device.bufferframes * obj.device.localbuffers]);

    if(FAILED(obj.pAudioClient->GetService(IID_IAudioRenderClient, (void**)&obj.pRenderClient))) {
      return false;
    }
    BYTE *endpointbuffer;
    while(FAILED(obj.pRenderClient->GetBuffer(obj.device.bufferframes, &endpointbuffer)));
    memset(endpointbuffer, 0, obj.device.bufferframes*4);
    obj.pRenderClient->ReleaseBuffer(obj.device.bufferframes, 0);
    obj.pAudioClient->Start();

    SetThreadPriority(obj.hThread, THREAD_PRIORITY_TIME_CRITICAL);
    HMODULE hModule = LoadLibrary(L"dwmapi.dll");
    if(hModule) {
      HRESULT WINAPI (*DwmEnableMMCSS)(BOOL) = (HRESULT WINAPI (*)(BOOL))GetProcAddress(hModule, "DwmEnableMMCSS");
      if(DwmEnableMMCSS) {
        DwmEnableMMCSS(TRUE);
      }
      FreeLibrary(hModule);
    }
    hModule = LoadLibrary(L"avrt.dll");
    if(hModule) {
      HANDLE WINAPI (*AvSetMmThreadCharacteristicsW)(LPCTSTR,LPDWORD) =
        (HANDLE WINAPI (*)(LPCTSTR,LPDWORD))GetProcAddress(hModule, "AvSetMmThreadCharacteristicsW");
      if(AvSetMmThreadCharacteristicsW) {
        DWORD taskindex = 0;
        HANDLE hTask = AvSetMmThreadCharacteristicsW(L"Pro Audio", &taskindex);
        BOOL WINAPI (*AvSetMmThreadPriority)(HANDLE,int) =
          (BOOL WINAPI (*)(HANDLE,int))GetProcAddress(hModule, "AvSetMmThreadPriority");
        if(AvSetMmThreadPriority) {
          AvSetMmThreadPriority(hTask, 2 /*AVRT_PRIORITY_CRITICAL*/);
        }
      }
      FreeLibrary(hModule);
    }
    timeBeginPeriod(1);

    return true;
  }

  static DWORD WINAPI ThreadProc(pAudioWASAPI& obj) {
    const HANDLE hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    if(!threadinit(obj, hEvent)) {
      InterlockedIncrement((volatile LONG*) &obj.Terminate);
      SetEvent(obj.hEvent);
      return false;
    }
    SetEvent(obj.hEvent);

    unsigned immediatebuffer = 0;
    //Set up local cache
    const unsigned bufferframes = obj.device.bufferframes;
    IAudioRenderClient *pRenderClient = obj.pRenderClient;
    volatile uint32_t const *buffersfilled = &obj.device.buffersfilled;
    const uint32_t *localbuffer = (const uint32_t*) obj.device.localbuffer;
    const HANDLE BufferAvailable = obj.BufferAvailable;
    const unsigned localbuffers = obj.device.localbuffers;

    while(true) {
      WaitForSingleObject(hEvent, INFINITE);
      BYTE *endpointbuffer;
      pRenderClient->GetBuffer(bufferframes, &endpointbuffer);
      if(*buffersfilled) {
        memcpy(endpointbuffer, localbuffer + immediatebuffer*bufferframes, bufferframes<<2);
        pRenderClient->ReleaseBuffer(bufferframes, 0);
        InterlockedDecrement((volatile LONG*) buffersfilled);
        SetEvent(BufferAvailable);
        immediatebuffer++; if(immediatebuffer == localbuffers) immediatebuffer = 0;
      } else {
        pRenderClient->ReleaseBuffer(bufferframes, AUDCLNT_BUFFERFLAGS_SILENT);
      }
    }

    threadterm(obj);
    return true;
  }

  static void threadterm(pAudioWASAPI& obj) {
    if(obj.pRenderClient) {
      if(obj.pAudioClient) {
        BYTE *endpointbuffer;
        obj.pRenderClient->GetBuffer(obj.device.bufferframes, &endpointbuffer);
        obj.pRenderClient->ReleaseBuffer(obj.device.bufferframes, AUDCLNT_BUFFERFLAGS_SILENT);
        obj.pAudioClient->Stop();
      }
      obj.pRenderClient->Release();
      obj.pRenderClient = 0;
    }
    if(obj.pAudioClient) {
      obj.pAudioClient->Release();
      obj.pAudioClient = 0;
    }
    if(obj.pDevice) {
      obj.pDevice->Release();
      obj.pDevice = 0;
    }
  }

  void term() {
    InterlockedIncrement((volatile LONG*) &Terminate);
    DWORD ExitCode;
    while(GetExitCodeThread(hThread, &ExitCode) && ExitCode == STILL_ACTIVE);
    CloseHandle(hThread);
    Terminate = 0;

    ResetEvent(hEvent);
    SetEvent(BufferAvailable);

    if(device.localbuffer) {
      delete[] device.localbuffer;
      device.localbuffer = 0;
    }
    device.buffersfilled = 0;
    device.onbuffer = 0;
    device.frameoffset = 0;
  }

  pAudioWASAPI() {
    pDevice = 0;
    pAudioClient = 0;
    pRenderClient = 0;

    hEvent = 0;
    BufferAvailable = 0;
    Terminate = 0;

    device.localbuffer = 0;
    device.buffersfilled = 0;
    device.onbuffer = 0;
    device.frameoffset = 0;
  }
};

DeclareAudio(WASAPI)

};